@@ -30,7 +30,13 @@
  */
 
 #include "file.h"
+
+#ifndef	lint
+FILE_RCSID("@(#)$File: apprentice.c,v 1.147 2009/02/03 20:27:51 christos Exp $")
+#endif	/* lint */
+
 #include "magic.h"
+#include "patchlevel.h"
 #include <stdlib.h>
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
@@ -39,15 +45,10 @@
 #include <assert.h>
 #include <ctype.h>
 #include <fcntl.h>
-#include <sys/stat.h>
-#include <sys/param.h>
 #ifdef QUICK
 #include <sys/mman.h>
 #endif
-
-#ifndef	lint
-FILE_RCSID("@(#)$File: apprentice.c,v 1.105 2007/05/16 20:51:40 christos Exp $")
-#endif	/* lint */
+#include <dirent.h>
 
 #define	EATAB {while (isascii((unsigned char) *l) && \
 		      isspace((unsigned char) *l))  ++l;}
@@ -96,24 +97,43 @@ private void eatsize(const char **);
 private int apprentice_1(struct magic_set *, const char *, int, struct mlist *);
 private size_t apprentice_magic_strength(const struct magic *);
 private int apprentice_sort(const void *, const void *);
-private int apprentice_file(struct magic_set *, struct magic **, uint32_t *,
+private int apprentice_load(struct magic_set *, struct magic **, uint32_t *,
     const char *, int);
 private void byteswap(struct magic *, uint32_t);
 private void bs1(struct magic *);
 private uint16_t swap2(uint16_t);
 private uint32_t swap4(uint32_t);
 private uint64_t swap8(uint64_t);
-private char *mkdbname(const char *, char *, size_t, int);
+private char *mkdbname(struct magic_set *, const char *, int);
 private int apprentice_map(struct magic_set *, struct magic **, uint32_t *,
     const char *);
 private int apprentice_compile(struct magic_set *, struct magic **, uint32_t *,
     const char *);
 private int check_format_type(const char *, int);
 private int check_format(struct magic_set *, struct magic *);
+private int get_op(char);
+private int parse_mime(struct magic_set *, struct magic_entry *, const char *);
+private int parse_strength(struct magic_set *, struct magic_entry *, const char *);
+private int parse_apple(struct magic_set *, struct magic_entry *, const char *);
+
 
 private size_t maxmagic = 0;
 private size_t magicsize = sizeof(struct magic);
 
+private const char usg_hdr[] = "cont\toffset\ttype\topcode\tmask\tvalue\tdesc";
+
+private struct {
+	const char *name;
+	size_t len;
+	int (*fun)(struct magic_set *, struct magic_entry *, const char *);
+} bang[] = {
+#define	DECLARE_FIELD(name) { # name, sizeof(# name) - 1, parse_ ## name }
+	DECLARE_FIELD(mime),
+	DECLARE_FIELD(apple),
+	DECLARE_FIELD(strength),
+#undef	DECLARE_FIELD
+	{ NULL, 0, NULL }
+};
 
 #ifdef COMPILE_ONLY
 
@@ -149,13 +169,13 @@ main(int argc, char *argv[])
 #endif /* COMPILE_ONLY */
 
 static const struct type_tbl_s {
-	const char *name;
+	const char name[16];
 	const size_t len;
 	const int type;
 	const int format;
 } type_tbl[] = {
 # define XX(s)		s, (sizeof(s) - 1)
-# define XX_NULL	NULL, 0
+# define XX_NULL	"", 0
 	{ XX("byte"),		FILE_BYTE,		FILE_FMT_NUM },
 	{ XX("short"),		FILE_SHORT,		FILE_FMT_NUM },
 	{ XX("default"),	FILE_DEFAULT,		FILE_FMT_STR },
@@ -188,6 +208,15 @@ static const struct type_tbl_s {
 	{ XX("qldate"),		FILE_QLDATE,		FILE_FMT_STR },
 	{ XX("leqldate"),	FILE_LEQLDATE,		FILE_FMT_STR },
 	{ XX("beqldate"),	FILE_BEQLDATE,		FILE_FMT_STR },
+	{ XX("float"),		FILE_FLOAT,		FILE_FMT_FLOAT },
+	{ XX("befloat"),	FILE_BEFLOAT,		FILE_FMT_FLOAT },
+	{ XX("lefloat"),	FILE_LEFLOAT,		FILE_FMT_FLOAT },
+	{ XX("double"),		FILE_DOUBLE,		FILE_FMT_DOUBLE },
+	{ XX("bedouble"),	FILE_BEDOUBLE,		FILE_FMT_DOUBLE },
+	{ XX("ledouble"),	FILE_LEDOUBLE,		FILE_FMT_DOUBLE },
+	{ XX("leid3"),		FILE_LEID3,		FILE_FMT_NUM },
+	{ XX("beid3"),		FILE_BEID3,		FILE_FMT_NUM },
+	{ XX("indirect"),	FILE_INDIRECT,		FILE_FMT_NONE },
 	{ XX_NULL,		FILE_INVALID,		FILE_FMT_NONE },
 # undef XX
 # undef XX_NULL
@@ -198,7 +227,7 @@ get_type(const char *l, const char **t)
 {
 	const struct type_tbl_s *p;
 
-	for (p = type_tbl; p->name; p++) {
+	for (p = type_tbl; p->len; p++) {
 		if (strncmp(l, p->name, p->len) == 0) {
 			if (t)
 				*t = l + p->len;
@@ -218,7 +247,7 @@ init_file_tables(void)
 		return;
 	done++;
 
-	for (p = type_tbl; p->name; p++) {
+	for (p = type_tbl; p->len; p++) {
 		assert(p->type < FILE_NAMES_SIZE);
 		file_names[p->type] = p->name;
 		file_formats[p->type] = p->format;
@@ -226,7 +255,7 @@ init_file_tables(void)
 }
 
 /*
- * Handle one file.
+ * Handle one file or directory.
  */
 private int
 apprentice_1(struct magic_set *ms, const char *fn, int action,
@@ -246,7 +275,7 @@ apprentice_1(struct magic_set *ms, const char *fn, int action,
 	}
 
 	if (action == FILE_COMPILE) {
-		rv = apprentice_file(ms, &magic, &nmagic, fn, action);
+		rv = apprentice_load(ms, &magic, &nmagic, fn, action);
 		if (rv != 0)
 			return -1;
 		rv = apprentice_compile(ms, &magic, &nmagic, fn);
@@ -258,20 +287,19 @@ apprentice_1(struct magic_set *ms, const char *fn, int action,
 	if ((rv = apprentice_map(ms, &magic, &nmagic, fn)) == -1) {
 		if (ms->flags & MAGIC_CHECK)
 			file_magwarn(ms, "using regular magic file `%s'", fn);
-		rv = apprentice_file(ms, &magic, &nmagic, fn, action);
+		rv = apprentice_load(ms, &magic, &nmagic, fn, action);
 		if (rv != 0)
 			return -1;
-		mapped = 0;
 	}
 
 	mapped = rv;
 	     
-	if (magic == NULL || nmagic == 0) {
+	if (magic == NULL) {
 		file_delmagic(magic, mapped, nmagic);
 		return -1;
 	}
 
-	if ((ml = malloc(sizeof(*ml))) == NULL) {
+	if ((ml = CAST(struct mlist *, malloc(sizeof(*ml)))) == NULL) {
 		file_delmagic(magic, mapped, nmagic);
 		file_oomem(ms, sizeof(*ml));
 		return -1;
@@ -296,10 +324,12 @@ file_delmagic(struct magic *p, int type, size_t entries)
 	if (p == NULL)
 		return;
 	switch (type) {
+#ifdef QUICK
 	case 2:
 		p--;
 		(void)munmap((void *)p, sizeof(*p) * (entries + 1));
 		break;
+#endif
 	case 1:
 		p--;
 		/*FALLTHROUGH*/
@@ -311,14 +341,13 @@ file_delmagic(struct magic *p, int type, size_t entries)
 	}
 }
 
-/* const char *fn: list of magic files */
+/* const char *fn: list of magic files and directories */
 protected struct mlist *
 file_apprentice(struct magic_set *ms, const char *fn, int action)
 {
-	char *p, *mfn, *afn = NULL;
+	char *p, *mfn;
 	int file_err, errs = -1;
 	struct mlist *mlist;
-	static const char mime[] = ".mime";
 
 	init_file_tables();
 
@@ -333,7 +362,7 @@ file_apprentice(struct magic_set *ms, const char *fn, int action)
 	}
 	fn = mfn;
 
-	if ((mlist = malloc(sizeof(*mlist))) == NULL) {
+	if ((mlist = CAST(struct mlist *, malloc(sizeof(*mlist)))) == NULL) {
 		free(mfn);
 		file_oomem(ms, sizeof(*mlist));
 		return NULL;
@@ -346,25 +375,8 @@ file_apprentice(struct magic_set *ms, const char *fn, int action)
 			*p++ = '\0';
 		if (*fn == '\0')
 			break;
-		if (ms->flags & MAGIC_MIME) {
-			size_t len = strlen(fn) + sizeof(mime);
-			if ((afn = malloc(len)) == NULL) {
-				free(mfn);
-				free(mlist);
-				file_oomem(ms, len);
-				return NULL;
-			}
-			(void)strcpy(afn, fn);
-			(void)strcat(afn, mime);
-			fn = afn;
-		}
 		file_err = apprentice_1(ms, fn, action, mlist);
-		if (file_err > errs)
-			errs = file_err;
-		if (afn) {
-			free(afn);
-			afn = NULL;
-		}
+		errs = MAX(errs, file_err);
 		fn = p;
 	}
 	if (errs == -1) {
@@ -389,6 +401,8 @@ apprentice_magic_strength(const struct magic *m)
 
 	switch (m->type) {
 	case FILE_DEFAULT:	/* make sure this sorts last */
+		if (m->factor_op != FILE_FACTOR_OP_NONE)
+			abort();
 		return 0;
 
 	case FILE_BYTE:
@@ -420,7 +434,7 @@ apprentice_magic_strength(const struct magic *m)
 
 	case FILE_SEARCH:
 	case FILE_REGEX:
-		val += m->vallen;
+		val += m->vallen * MAX(MULT / m->vallen, 1);
 		break;
 
 	case FILE_DATE:
@@ -431,6 +445,9 @@ apprentice_magic_strength(const struct magic *m)
 	case FILE_LELDATE:
 	case FILE_BELDATE:
 	case FILE_MELDATE:
+	case FILE_FLOAT:
+	case FILE_BEFLOAT:
+	case FILE_LEFLOAT:
 		val += 4 * MULT;
 		break;
 
@@ -443,6 +460,9 @@ apprentice_magic_strength(const struct magic *m)
 	case FILE_QLDATE:
 	case FILE_LEQLDATE:
 	case FILE_BEQLDATE:
+	case FILE_DOUBLE:
+	case FILE_BEDOUBLE:
+	case FILE_LEDOUBLE:
 		val += 8 * MULT;
 		break;
 
@@ -454,10 +474,10 @@ apprentice_magic_strength(const struct magic *m)
 
 	switch (m->reln) {
 	case 'x':	/* matches anything penalize */
+	case '!':       /* matches almost anything penalize */
 		val = 0;
 		break;
 
-	case '!':
 	case '=':	/* Exact match, prefer */
 		val += MULT;
 		break;
@@ -480,6 +500,31 @@ apprentice_magic_strength(const struct magic *m)
 	if (val == 0)	/* ensure we only return 0 for FILE_DEFAULT */
 		val = 1;
 
+	switch (m->factor_op) {
+	case FILE_FACTOR_OP_NONE:
+		break;
+	case FILE_FACTOR_OP_PLUS:
+		val += m->factor;
+		break;
+	case FILE_FACTOR_OP_MINUS:
+		val -= m->factor;
+		break;
+	case FILE_FACTOR_OP_TIMES:
+		val *= m->factor;
+		break;
+	case FILE_FACTOR_OP_DIV:
+		val /= m->factor;
+		break;
+	default:
+		abort();
+	}
+
+	/*
+	 * Magic entries with no description get a bonus because they depend
+	 * on subsequent magic entries to print something.
+	 */
+	if (m->desc[0] == '\0')
+		val++;
 	return val;
 }
 
@@ -489,8 +534,8 @@ apprentice_magic_strength(const struct magic *m)
 private int
 apprentice_sort(const void *a, const void *b)
 {
-	const struct magic_entry *ma = a;
-	const struct magic_entry *mb = b;
+	const struct magic_entry *ma = CAST(const struct magic_entry *, a);
+	const struct magic_entry *mb = CAST(const struct magic_entry *, b);
 	size_t sa = apprentice_magic_strength(ma->mp);
 	size_t sb = apprentice_magic_strength(mb->mp);
 	if (sa == sb)
@@ -501,36 +546,157 @@ apprentice_sort(const void *a, const void *b)
 		return 1;
 }
 
-/*
- * parse from a file
- * const char *fn: name of magic file
- */
-private int
-apprentice_file(struct magic_set *ms, struct magic **magicp, uint32_t *nmagicp,
-    const char *fn, int action)
+private void
+set_test_type(struct magic *mstart, struct magic *m)
 {
-	private const char hdr[] =
-		"cont\toffset\ttype\topcode\tmask\tvalue\tdesc";
-	FILE *f;
-	char line[BUFSIZ+1];
-	int errs = 0;
-	struct magic_entry *marray;
-	uint32_t marraycount, i, mentrycount = 0;
+	switch (m->type) {
+	case FILE_BYTE:
+	case FILE_SHORT:
+	case FILE_LONG:
+	case FILE_DATE:
+	case FILE_BESHORT:
+	case FILE_BELONG:
+	case FILE_BEDATE:
+	case FILE_LESHORT:
+	case FILE_LELONG:
+	case FILE_LEDATE:
+	case FILE_LDATE:
+	case FILE_BELDATE:
+	case FILE_LELDATE:
+	case FILE_MEDATE:
+	case FILE_MELDATE:
+	case FILE_MELONG:
+	case FILE_QUAD:
+	case FILE_LEQUAD:
+	case FILE_BEQUAD:
+	case FILE_QDATE:
+	case FILE_LEQDATE:
+	case FILE_BEQDATE:
+	case FILE_QLDATE:
+	case FILE_LEQLDATE:
+	case FILE_BEQLDATE:
+	case FILE_FLOAT:
+	case FILE_BEFLOAT:
+	case FILE_LEFLOAT:
+	case FILE_DOUBLE:
+	case FILE_BEDOUBLE:
+	case FILE_LEDOUBLE:
+	case FILE_STRING:
+	case FILE_PSTRING:
+	case FILE_BESTRING16:
+	case FILE_LESTRING16:
+		/* binary test, set flag */
+		mstart->flag |= BINTEST;
+		break;
+	case FILE_REGEX:
+	case FILE_SEARCH:
+		/* binary test if pattern is not text */
+		if (file_looks_utf8(m->value.us, (size_t)m->vallen, NULL,
+		    NULL) <= 0)
+			mstart->flag |= BINTEST;
+		break;
+	case FILE_DEFAULT:
+		/* can't deduce anything; we shouldn't see this at the
+		   top level anyway */
+		break;
+	case FILE_INVALID:
+	default:
+		/* invalid search type, but no need to complain here */
+		break;
+	}
+}
+
+/*
+ * Load and parse one file.
+ */
+private void
+load_1(struct magic_set *ms, int action, const char *fn, int *errs,
+   struct magic_entry **marray, uint32_t *marraycount)
+{
+	char line[BUFSIZ];
 	size_t lineno = 0;
-
-	ms->flags |= MAGIC_CHECK;	/* Enable checks for parsed files */
-
-	f = fopen(ms->file = fn, "r");
+	FILE *f = fopen(ms->file = fn, "r");
 	if (f == NULL) {
 		if (errno != ENOENT)
 			file_error(ms, errno, "cannot read magic file `%s'",
-			    fn);
-		return -1;
+				   fn);
+		(*errs)++;
+	} else {
+		/* read and parse this file */
+		for (ms->line = 1; fgets(line, sizeof(line), f) != NULL; ms->line++) {
+			size_t len;
+			len = strlen(line);
+			if (len == 0) /* null line, garbage, etc */
+				continue;
+			if (line[len - 1] == '\n') {
+				lineno++;
+				line[len - 1] = '\0'; /* delete newline */
+			}
+			if (line[0] == '\0')	/* empty, do not parse */
+				continue;
+			if (line[0] == '#')	/* comment, do not parse */
+				continue;
+			if (line[0] == '!' && line[1] == ':') {
+				size_t i;
+
+				for (i = 0; bang[i].name != NULL; i++) {
+					if (len - 2 > bang[i].len &&
+					    memcmp(bang[i].name, line + 2,
+					    bang[i].len) == 0)
+						break;
+				}
+				if (bang[i].name == NULL) {
+					file_error(ms, 0,
+					    "Unknown !: entry `%s'", line);
+					(*errs)++;
+					continue;
+				}
+				if (*marraycount == 0) {
+					file_error(ms, 0,
+					    "No current entry for :!%s type",
+						bang[i].name);
+					(*errs)++;
+					continue;
+				}
+				if ((*bang[i].fun)(ms, 
+				    &(*marray)[*marraycount - 1],
+				    line + bang[i].len + 2) != 0) {
+					(*errs)++;
+					continue;
+				}
+				continue;
+			}
+			if (parse(ms, marray, marraycount, line, lineno,
+			    action) != 0)
+				(*errs)++;
+		}
+
+		(void)fclose(f);
 	}
+}
+
+/*
+ * parse a file or directory of files
+ * const char *fn: name of magic file or directory
+ */
+private int
+apprentice_load(struct magic_set *ms, struct magic **magicp, uint32_t *nmagicp,
+    const char *fn, int action)
+{
+	int errs = 0;
+	struct magic_entry *marray;
+	uint32_t marraycount, i, mentrycount = 0, starttest;
+	size_t slen;
+	char subfn[MAXPATHLEN];
+	struct stat st;
+	DIR *dir;
+	struct dirent *d;
+
+	ms->flags |= MAGIC_CHECK;	/* Enable checks for parsed files */
 
         maxmagic = MAXMAGIS;
-	if ((marray = calloc(maxmagic, sizeof(*marray))) == NULL) {
-		(void)fclose(f);
+	if ((marray = CAST(struct magic_entry *, calloc(maxmagic,
+	    sizeof(*marray)))) == NULL) {
 		file_oomem(ms, maxmagic * sizeof(*marray));
 		return -1;
 	}
@@ -538,32 +704,67 @@ apprentice_file(struct magic_set *ms, struct magic **magicp, uint32_t *nmagicp,
 
 	/* print silly verbose header for USG compat. */
 	if (action == FILE_CHECK)
-		(void)fprintf(stderr, "%s\n", hdr);
+		(void)fprintf(stderr, "%s\n", usg_hdr);
 
-	/* read and parse this file */
-	for (ms->line = 1; fgets(line, BUFSIZ, f) != NULL; ms->line++) {
-		size_t len;
-		len = strlen(line);
-		if (len == 0) /* null line, garbage, etc */
-			continue;
-		if (line[len - 1] == '\n') {
-			lineno++;
-			line[len - 1] = '\0'; /* delete newline */
-		}
-		if (line[0] == '\0')	/* empty, do not parse */
-			continue;
-		if (line[0] == '#')	/* comment, do not parse */
-			continue;
-		if (parse(ms, &marray, &marraycount, line, lineno, action) != 0)
+	/* load directory or file */
+        /* FIXME: Read file names and sort them to prevent
+           non-determinism. See Debian bug #488562. */
+	if (stat(fn, &st) == 0 && S_ISDIR(st.st_mode)) {
+		dir = opendir(fn);
+		if (dir) {
+			while ((d = readdir(dir)) != NULL) {
+				snprintf(subfn, sizeof(subfn), "%s/%s",
+				    fn, d->d_name);
+				if (stat(subfn, &st) == 0 &&
+				    S_ISREG(st.st_mode)) {
+					load_1(ms, action, subfn, &errs,
+					    &marray, &marraycount);
+				}
+			}
+			closedir(dir);
+		} else
 			errs++;
-	}
-
-	(void)fclose(f);
+	} else
+		load_1(ms, action, fn, &errs, &marray, &marraycount);
 	if (errs)
 		goto out;
 
-#ifndef NOORDER
+	/* Set types of tests */
+	for (i = 0; i < marraycount; ) {
+		if (marray[i].mp->cont_level != 0) {
+			i++;
+			continue;
+		}
+
+		starttest = i;
+		do {
+			static const char text[] = "text";
+			static const char binary[] = "binary";
+			static const size_t len = sizeof(text);
+			set_test_type(marray[starttest].mp, marray[i].mp);
+			if ((ms->flags & MAGIC_DEBUG) == 0)
+				continue;
+			(void)fprintf(stderr, "%s%s%s: %s\n",
+			    marray[i].mp->mimetype,
+			    marray[i].mp->mimetype[0] == '\0' ? "" : "; ",
+			    marray[i].mp->desc[0] ? marray[i].mp->desc :
+			    "(no description)",
+			    marray[i].mp->flag & BINTEST ? binary : text);
+			if (marray[i].mp->flag & BINTEST) {
+				char *p = strstr(marray[i].mp->desc, text);
+				if (p && (p == marray[i].mp->desc ||
+				    isspace((unsigned char)p[-1])) &&
+				    (p + len - marray[i].mp->desc == 
+				    MAXstring || (p[len] == '\0' ||
+				    isspace((unsigned char)p[len]))))
+					(void)fprintf(stderr, "*** Possible "
+					    "binary test for text type\n");
+			}
+		} while (++i < marraycount && marray[i].mp->cont_level != 0);
+	}
+
 	qsort(marray, marraycount, sizeof(*marray), apprentice_sort);
+
 	/*
 	 * Make sure that any level 0 "default" line is last (if one exists).
 	 */
@@ -581,13 +782,13 @@ apprentice_file(struct magic_set *ms, struct magic **magicp, uint32_t *nmagicp,
 			break;					    
 		}
 	}
-#endif
 
 	for (i = 0; i < marraycount; i++)
 		mentrycount += marray[i].cont_count;
 
-	if ((*magicp = malloc(sizeof(**magicp) * mentrycount)) == NULL) {
-		file_oomem(ms, sizeof(**magicp) * mentrycount);
+	slen = sizeof(**magicp) * mentrycount;
+	if ((*magicp = CAST(struct magic *, malloc(slen))) == NULL) {
+		file_oomem(ms, slen);
 		errs++;
 		goto out;
 	}
@@ -646,6 +847,9 @@ file_signextend(struct magic_set *ms, struct magic *m, uint64_t v)
 		case FILE_BELONG:
 		case FILE_LELONG:
 		case FILE_MELONG:
+		case FILE_FLOAT:
+		case FILE_BEFLOAT:
+		case FILE_LEFLOAT:
 			v = (int32_t) v;
 			break;
 		case FILE_QUAD:
@@ -657,6 +861,9 @@ file_signextend(struct magic_set *ms, struct magic *m, uint64_t v)
 		case FILE_BEQLDATE:
 		case FILE_LEQDATE:
 		case FILE_LEQLDATE:
+		case FILE_DOUBLE:
+		case FILE_BEDOUBLE:
+		case FILE_LEDOUBLE:
 			v = (int64_t) v;
 			break;
 		case FILE_STRING:
@@ -666,6 +873,7 @@ file_signextend(struct magic_set *ms, struct magic *m, uint64_t v)
 		case FILE_REGEX:
 		case FILE_SEARCH:
 		case FILE_DEFAULT:
+		case FILE_INDIRECT:
 			break;
 		default:
 			if (ms->flags & MAGIC_CHECK)
@@ -678,7 +886,7 @@ file_signextend(struct magic_set *ms, struct magic *m, uint64_t v)
 }
 
 private int
-string_modifier_check(struct magic_set *ms, struct magic const *m)
+string_modifier_check(struct magic_set *ms, struct magic *m)
 {
 	if ((ms->flags & MAGIC_CHECK) == 0)
 		return 0;
@@ -687,19 +895,28 @@ string_modifier_check(struct magic_set *ms, struct magic const *m)
 	case FILE_BESTRING16:
 	case FILE_LESTRING16:
 		if (m->str_flags != 0) {
-			file_magwarn(ms, "no modifiers allowed for 16-bit strings\n");
+			file_magwarn(ms,
+			    "no modifiers allowed for 16-bit strings\n");
 			return -1;
 		}
 		break;
 	case FILE_STRING:
 	case FILE_PSTRING:
 		if ((m->str_flags & REGEX_OFFSET_START) != 0) {
-			file_magwarn(ms, "'/%c' only allowed on regex and search\n",
+			file_magwarn(ms,
+			    "'/%c' only allowed on regex and search\n",
 			    CHAR_REGEX_OFFSET_START);
 			return -1;
 		}
 		break;
 	case FILE_SEARCH:
+		if (m->str_range == 0) {
+			file_magwarn(ms,
+			    "missing range; defaulting to %d\n",
+                            STRING_DEFAULT_RANGE);
+			m->str_range = STRING_DEFAULT_RANGE;
+			return -1;
+		}
 		break;
 	case FILE_REGEX:
 		if ((m->str_flags & STRING_COMPACT_BLANK) != 0) {
@@ -750,19 +967,19 @@ get_op(char c)
 private int
 get_cond(const char *l, const char **t)
 {
-	static struct cond_tbl_s {
-		const char *name;
-		const size_t len;
-		const int cond;
+	static const struct cond_tbl_s {
+		char name[8];
+		size_t len;
+		int cond;
 	} cond_tbl[] = {
 		{ "if",		2,	COND_IF },
 		{ "elif",	4,	COND_ELIF },
 		{ "else",	4,	COND_ELSE },
-		{ NULL, 	0,	COND_NONE },
+		{ "",		0,	COND_NONE },
 	};
-	struct cond_tbl_s *p;
+	const struct cond_tbl_s *p;
 
-	for (p = cond_tbl; p->name; p++) {
+	for (p = cond_tbl; p->len; p++) {
 		if (strncmp(l, p->name, p->len) == 0 &&
 		    isspace((unsigned char)l[p->len])) {
 			if (t)
@@ -860,7 +1077,8 @@ parse(struct magic_set *ms, struct magic_entry **mentryp, uint32_t *nmentryp,
 		if (me->cont_count == me->max_count) {
 			struct magic *nm;
 			size_t cnt = me->max_count + ALLOC_CHUNK;
-			if ((nm = realloc(me->mp, sizeof(*nm) * cnt)) == NULL) {
+			if ((nm = CAST(struct magic *, realloc(me->mp,
+			    sizeof(*nm) * cnt))) == NULL) {
 				file_oomem(ms, sizeof(*nm) * cnt);
 				return -1;
 			}
@@ -875,7 +1093,8 @@ parse(struct magic_set *ms, struct magic_entry **mentryp, uint32_t *nmentryp,
 			struct magic_entry *mp;
 
 			maxmagic += ALLOC_INCR;
-			if ((mp = realloc(*mentryp, sizeof(*mp) * maxmagic)) ==
+			if ((mp = CAST(struct magic_entry *,
+			    realloc(*mentryp, sizeof(*mp) * maxmagic))) ==
 			    NULL) {
 				file_oomem(ms, sizeof(*mp) * maxmagic);
 				return -1;
@@ -886,8 +1105,9 @@ parse(struct magic_set *ms, struct magic_entry **mentryp, uint32_t *nmentryp,
 		}
 		me = &(*mentryp)[*nmentryp];
 		if (me->mp == NULL) {
-			if ((m = malloc(sizeof(*m) * ALLOC_CHUNK)) == NULL) {
-				file_oomem(ms, sizeof(*m) * ALLOC_CHUNK);
+			size_t len = sizeof(*m) * ALLOC_CHUNK;
+			if ((m = CAST(struct magic *, malloc(len))) == NULL) {
+				file_oomem(ms, len);
 				return -1;
 			}
 			me->mp = m;
@@ -895,6 +1115,7 @@ parse(struct magic_set *ms, struct magic_entry **mentryp, uint32_t *nmentryp,
 		} else
 			m = me->mp;
 		(void)memset(m, 0, sizeof(*m));
+		m->factor_op = FILE_FACTOR_OP_NONE;
 		m->cont_level = 0;
 		me->cont_count = 1;
 	}
@@ -959,6 +1180,22 @@ parse(struct magic_set *ms, struct magic_entry **mentryp, uint32_t *nmentryp,
 			case 'B':
 				m->in_type = FILE_BYTE;
 				break;
+			case 'e':
+			case 'f':
+			case 'g':
+				m->in_type = FILE_LEDOUBLE;
+				break;
+			case 'E':
+			case 'F':
+			case 'G':
+				m->in_type = FILE_BEDOUBLE;
+				break;
+			case 'i':
+				m->in_type = FILE_LEID3;
+				break;
+			case 'I':
+				m->in_type = FILE_BEID3;
+				break;
 			default:
 				if (ms->flags & MAGIC_CHECK)
 					file_magwarn(ms,
@@ -1029,7 +1266,7 @@ parse(struct magic_set *ms, struct magic_entry **mentryp, uint32_t *nmentryp,
 			file_magwarn(ms, "'~' invalid for string types");
 		++l;
 	}
-	m->str_count = 0;
+	m->str_range = 0;
 	m->str_flags = 0;
 	m->num_mask = 0;
 	if ((op = get_op(*l)) != -1) {
@@ -1043,22 +1280,24 @@ parse(struct magic_set *ms, struct magic_entry **mentryp, uint32_t *nmentryp,
 			eatsize(&l);
 		}
 		else if (op == FILE_OPDIVIDE) {
-			int have_count = 0;
+			int have_range = 0;
 			while (!isspace((unsigned char)*++l)) {
 				switch (*l) {
-				/* for portability avoid "case '0' ... '9':" */
 				case '0':  case '1':  case '2':
 				case '3':  case '4':  case '5':
 				case '6':  case '7':  case '8':
-				case '9': {
-					if (have_count && ms->flags & MAGIC_CHECK)
+				case '9':
+					if (have_range &&
+					    (ms->flags & MAGIC_CHECK))
 						file_magwarn(ms,
-						    "multiple counts");
-					have_count = 1;
-					m->str_count = strtoul(l, &t, 0);
+						    "multiple ranges");
+					have_range = 1;
+					m->str_range = strtoul(l, &t, 0);
+					if (m->str_range == 0)
+						file_magwarn(ms,
+						    "zero range");
 					l = t - 1;
 					break;
-				}
 				case CHAR_COMPACT_BLANK:
 					m->str_flags |= STRING_COMPACT_BLANK;
 					break;
@@ -1083,7 +1322,8 @@ parse(struct magic_set *ms, struct magic_entry **mentryp, uint32_t *nmentryp,
 					return -1;
 				}
 				/* allow multiple '/' for readability */
-				if (l[1] == '/' && !isspace((unsigned char)l[2]))
+				if (l[1] == '/' &&
+				    !isspace((unsigned char)l[2]))
 					l++;
 			}
 			if (string_modifier_check(ms, m) == -1)
@@ -1104,6 +1344,17 @@ parse(struct magic_set *ms, struct magic_entry **mentryp, uint32_t *nmentryp,
 	switch (*l) {
 	case '>':
 	case '<':
+  		m->reln = *l;
+  		++l;
+		if (*l == '=') {
+			if (ms->flags & MAGIC_CHECK) {
+				file_magwarn(ms, "%c= not supported",
+				    m->reln);
+				return -1;
+			}
+		   ++l;
+		}
+		break;
 	/* Old-style anding: "0 byte &0x80 dynamically linked" */
 	case '&':
 	case '^':
@@ -1146,13 +1397,12 @@ parse(struct magic_set *ms, struct magic_entry **mentryp, uint32_t *nmentryp,
 	EATAB;
 	if (l[0] == '\b') {
 		++l;
-		m->nospflag = 1;
+		m->flag |= NOSPACE;
 	} else if ((l[0] == '\\') && (l[1] == 'b')) {
 		++l;
 		++l;
-		m->nospflag = 1;
-	} else
-		m->nospflag = 0;
+		m->flag |= NOSPACE;
+	}
 	for (i = 0; (m->desc[i++] = *l++) != '\0' && i < sizeof(m->desc); )
 		continue;
 	if (i == sizeof(m->desc)) {
@@ -1174,11 +1424,133 @@ parse(struct magic_set *ms, struct magic_entry **mentryp, uint32_t *nmentryp,
 		file_mdump(m);
 	}
 #endif
+	m->mimetype[0] = '\0';		/* initialise MIME type to none */
 	if (m->cont_level == 0)
 		++(*nmentryp);		/* make room for next */
 	return 0;
 }
 
+/*
+ * parse a STRENGTH annotation line from magic file, put into magic[index - 1]
+ * if valid
+ */
+private int
+parse_strength(struct magic_set *ms, struct magic_entry *me, const char *line)
+{
+	const char *l = line;
+	char *el;
+	unsigned long factor;
+	struct magic *m = &me->mp[0];
+
+	if (m->factor_op != FILE_FACTOR_OP_NONE) {
+		file_magwarn(ms,
+		    "Current entry already has a strength type: %c %d",
+		    m->factor_op, m->factor);
+		return -1;
+	}
+	EATAB;
+	switch (*l) {
+	case FILE_FACTOR_OP_NONE:
+	case FILE_FACTOR_OP_PLUS:
+	case FILE_FACTOR_OP_MINUS:
+	case FILE_FACTOR_OP_TIMES:
+	case FILE_FACTOR_OP_DIV:
+		m->factor_op = *l++;
+		break;
+	default:
+		file_magwarn(ms, "Unknown factor op `%c'", *l);
+		return -1;
+	}
+	EATAB;
+	factor = strtoul(l, &el, 0);
+	if (factor > 255) {
+		file_magwarn(ms, "Too large factor `%lu'", factor);
+		goto out;
+	}
+	if (*el && !isspace((unsigned char)*el)) {
+		file_magwarn(ms, "Bad factor `%s'", l);
+		goto out;
+	}
+	m->factor = (uint8_t)factor;
+	if (m->factor == 0 && m->factor_op == FILE_FACTOR_OP_DIV) {
+		file_magwarn(ms, "Cannot have factor op `%c' and factor %u",
+		    m->factor_op, m->factor);
+		goto out;
+	}
+	return 0;
+out:
+	m->factor_op = FILE_FACTOR_OP_NONE;
+	m->factor = 0;
+	return -1;
+}
+
+/*
+ * Parse an Apple CREATOR/TYPE annotation from magic file and put it into magic[index - 1]
+ */
+private int
+parse_apple(struct magic_set *ms, struct magic_entry *me, const char *line)
+{
+	size_t i;
+	const char *l = line;
+	struct magic *m = &me->mp[me->cont_count == 0 ? 0 : me->cont_count - 1];
+
+	if (m->apple[0] != '\0') {
+		file_magwarn(ms, "Current entry already has a APPLE type `%.8s',"
+		    " new type `%s'", m->mimetype, l);
+		return -1;
+	}	
+
+	EATAB;
+	for (i = 0; *l && ((isascii((unsigned char)*l) && isalnum((unsigned char)*l))
+	     || strchr("-+/.", *l)) && i < sizeof(m->apple); m->apple[i++] = *l++)
+		continue;
+	if (i == sizeof(m->apple) && *l) {
+		if (ms->flags & MAGIC_CHECK)
+			file_magwarn(ms, "APPLE type `%s' truncated %zu",
+			    line, i);
+	}
+
+	if (i > 0)
+		return 0;
+	else
+		return -1;
+}
+
+/*
+ * parse a MIME annotation line from magic file, put into magic[index - 1]
+ * if valid
+ */
+private int
+parse_mime(struct magic_set *ms, struct magic_entry *me, const char *line)
+{
+	size_t i;
+	const char *l = line;
+	struct magic *m = &me->mp[me->cont_count == 0 ? 0 : me->cont_count - 1];
+
+	if (m->mimetype[0] != '\0') {
+		file_magwarn(ms, "Current entry already has a MIME type `%s',"
+		    " new type `%s'", m->mimetype, l);
+		return -1;
+	}	
+
+	EATAB;
+	for (i = 0; *l && ((isascii((unsigned char)*l) && isalnum((unsigned char)*l))
+	     || strchr("-+/.", *l)) && i < sizeof(m->mimetype); m->mimetype[i++] = *l++)
+		continue;
+	if (i == sizeof(m->mimetype)) {
+		m->desc[sizeof(m->mimetype) - 1] = '\0';
+		if (ms->flags & MAGIC_CHECK)
+			file_magwarn(ms, "MIME type `%s' truncated %zu",
+			    m->mimetype, i);
+	} else
+		m->mimetype[i] = '\0';
+
+	if (i > 0)
+		return 0;
+	else
+		return -1;
+}
+
 private int
 check_format_type(const char *ptr, int type)
 {
@@ -1252,6 +1624,31 @@ check_format_type(const char *ptr, int type)
 			return -1;
 		}
 		
+	case FILE_FMT_FLOAT:
+	case FILE_FMT_DOUBLE:
+		if (*ptr == '-')
+			ptr++;
+		if (*ptr == '.')
+			ptr++;
+		while (isdigit((unsigned char)*ptr)) ptr++;
+		if (*ptr == '.')
+			ptr++;
+		while (isdigit((unsigned char)*ptr)) ptr++;
+	
+		switch (*ptr++) {
+		case 'e':
+		case 'E':
+		case 'f':
+		case 'F':
+		case 'g':
+		case 'G':
+			return 0;
+			
+		default:
+			return -1;
+		}
+		
+
 	case FILE_FMT_STR:
 		if (*ptr == '-')
 			ptr++;
@@ -1298,12 +1695,12 @@ check_format(struct magic_set *ms, struct magic *m)
 	assert(file_nformats == file_nnames);
 
 	if (m->type >= file_nformats) {
-		file_error(ms, 0, "Internal error inconsistency between "
+		file_magwarn(ms, "Internal error inconsistency between "
 		    "m->type and format strings");		
 		return -1;
 	}
 	if (file_formats[m->type] == FILE_FMT_NONE) {
-		file_error(ms, 0, "No format string for `%s' with description "
+		file_magwarn(ms, "No format string for `%s' with description "
 		    "`%s'", m->desc, file_names[m->type]);
 		return -1;
 	}
@@ -1314,15 +1711,16 @@ check_format(struct magic_set *ms, struct magic *m)
 		 * TODO: this error message is unhelpful if the format
 		 * string is not one character long
 		 */
-		file_error(ms, 0, "Printf format `%c' is not valid for type "
-		    " `%s' in description `%s'", *ptr,
+		file_magwarn(ms, "Printf format `%c' is not valid for type "
+		    "`%s' in description `%s'",
+		    ptr && *ptr ? *ptr : '?',
 		    file_names[m->type], m->desc);
 		return -1;
 	}
 	
 	for (; *ptr; ptr++) {
 		if (*ptr == '%') {
-			file_error(ms, 0,
+			file_magwarn(ms,
 			    "Too many format strings (should have at most one) "
 			    "for `%s' with description `%s'",
 			    file_names[m->type], m->desc);
@@ -1357,6 +1755,30 @@ getvalue(struct magic_set *ms, struct magic *m, const char **p, int action)
 			return -1;
 		}
 		m->vallen = slen;
+		if (m->type == FILE_PSTRING)
+			m->vallen++;
+		return 0;
+	case FILE_FLOAT:
+	case FILE_BEFLOAT:
+	case FILE_LEFLOAT:
+		if (m->reln != 'x') {
+			char *ep;
+#ifdef HAVE_STRTOF
+			m->value.f = strtof(*p, &ep);
+#else
+			m->value.f = (float)strtod(*p, &ep);
+#endif
+			*p = ep;
+		}
+		return 0;
+	case FILE_DOUBLE:
+	case FILE_BEDOUBLE:
+	case FILE_LEDOUBLE:
+		if (m->reln != 'x') {
+			char *ep;
+			m->value.d = strtod(*p, &ep);
+			*p = ep;
+		}
 		return 0;
 	default:
 		if (m->reln != 'x') {
@@ -1631,51 +2053,51 @@ apprentice_map(struct magic_set *ms, struct magic **magicp, uint32_t *nmagicp,
 	uint32_t *ptr;
 	uint32_t version;
 	int needsbyteswap;
-	char buf[MAXPATHLEN];
-	char *dbname = mkdbname(fn, buf, sizeof(buf), 0);
+	char *dbname = NULL;
 	void *mm = NULL;
 
+	dbname = mkdbname(ms, fn, 0);
 	if (dbname == NULL)
-		return -1;
+		goto error2;
 
 	if ((fd = open(dbname, O_RDONLY|O_BINARY)) == -1)
-		return -1;
+		goto error2;
 
 	if (fstat(fd, &st) == -1) {
 		file_error(ms, errno, "cannot stat `%s'", dbname);
-		goto error;
+		goto error1;
 	}
-	if (st.st_size < 16) {
+	if (st.st_size < 8) {
 		file_error(ms, 0, "file `%s' is too small", dbname);
-		goto error;
+		goto error1;
 	}
 
 #ifdef QUICK
 	if ((mm = mmap(0, (size_t)st.st_size, PROT_READ|PROT_WRITE,
 	    MAP_PRIVATE|MAP_FILE, fd, (off_t)0)) == MAP_FAILED) {
 		file_error(ms, errno, "cannot map `%s'", dbname);
-		goto error;
+		goto error1;
 	}
 #define RET	2
 #else
-	if ((mm = malloc((size_t)st.st_size)) == NULL) {
+	if ((mm = CAST(void *, malloc((size_t)st.st_size))) == NULL) {
 		file_oomem(ms, (size_t)st.st_size);
-		goto error;
+		goto error1;
 	}
 	if (read(fd, mm, (size_t)st.st_size) != (size_t)st.st_size) {
 		file_badread(ms);
-		goto error;
+		goto error1;
 	}
 #define RET	1
 #endif
-	*magicp = mm;
+	*magicp = CAST(struct magic *, mm);
 	(void)close(fd);
 	fd = -1;
 	ptr = (uint32_t *)(void *)*magicp;
 	if (*ptr != MAGICNO) {
 		if (swap4(*ptr) != MAGICNO) {
-			file_error(ms, 0, "bad magic in `%s'");
-			goto error;
+			file_error(ms, 0, "bad magic in `%s'", dbname);
+			goto error1;
 		}
 		needsbyteswap = 1;
 	} else
@@ -1685,17 +2107,21 @@ apprentice_map(struct magic_set *ms, struct magic **magicp, uint32_t *nmagicp,
 	else
 		version = ptr[1];
 	if (version != VERSIONNO) {
-		file_error(ms, 0, "version mismatch (%d != %d) in `%s'",
-		    version, VERSIONNO, dbname);
-		goto error;
+		file_error(ms, 0, "File %d.%d supports only %d version magic "
+		    "files. `%s' is version %d", FILE_VERSION_MAJOR, patchlevel,
+		    VERSIONNO, dbname, version);
+		goto error1;
 	}
-	*nmagicp = (uint32_t)(st.st_size / sizeof(struct magic)) - 1;
+	*nmagicp = (uint32_t)(st.st_size / sizeof(struct magic));
+	if (*nmagicp > 0)
+		(*nmagicp)--;
 	(*magicp)++;
 	if (needsbyteswap)
 		byteswap(*magicp, *nmagicp);
+	free(dbname);
 	return RET;
 
-error:
+error1:
 	if (fd != -1)
 		(void)close(fd);
 	if (mm) {
@@ -1708,6 +2134,8 @@ error:
 		*magicp = NULL;
 		*nmagicp = 0;
 	}
+error2:
+	free(dbname);
 	return -1;
 }
 
@@ -1722,36 +2150,41 @@ apprentice_compile(struct magic_set *ms, struct magic **magicp,
     uint32_t *nmagicp, const char *fn)
 {
 	int fd;
-	char buf[MAXPATHLEN];
-	char *dbname = mkdbname(fn, buf, sizeof(buf), 1);
+	char *dbname;
+	int rv = -1;
+
+	dbname = mkdbname(ms, fn, 1);
 
 	if (dbname == NULL) 
-		return -1;
+		goto out;
 
 	if ((fd = open(dbname, O_WRONLY|O_CREAT|O_TRUNC|O_BINARY, 0644)) == -1) {
 		file_error(ms, errno, "cannot open `%s'", dbname);
-		return -1;
+		goto out;
 	}
 
 	if (write(fd, ar, sizeof(ar)) != (ssize_t)sizeof(ar)) {
 		file_error(ms, errno, "error writing `%s'", dbname);
-		return -1;
+		goto out;
 	}
 
 	if (lseek(fd, (off_t)sizeof(struct magic), SEEK_SET)
 	    != sizeof(struct magic)) {
 		file_error(ms, errno, "error seeking `%s'", dbname);
-		return -1;
+		goto out;
 	}
 
 	if (write(fd, *magicp, (sizeof(struct magic) * *nmagicp)) 
 	    != (ssize_t)(sizeof(struct magic) * *nmagicp)) {
 		file_error(ms, errno, "error writing `%s'", dbname);
-		return -1;
+		goto out;
 	}
 
 	(void)close(fd);
-	return 0;
+	rv = 0;
+out:
+	free(dbname);
+	return rv;
 }
 
 private const char ext[] = ".mgc";
@@ -1759,15 +2192,43 @@ private const char ext[] = ".mgc";
  * make a dbname
  */
 private char *
-mkdbname(const char *fn, char *buf, size_t bufsiz, int strip)
+mkdbname(struct magic_set *ms, const char *fn, int strip)
 {
+	const char *p, *q;
+	char *buf;
+
 	if (strip) {
-		const char *p;
 		if ((p = strrchr(fn, '/')) != NULL)
 			fn = ++p;
 	}
 
-	(void)snprintf(buf, bufsiz, "%s%s", fn, ext);
+	for (q = fn; *q; q++)
+		continue;
+	/* Look for .mgc */
+	for (p = ext + sizeof(ext) - 1; p >= ext && q >= fn; p--, q--)
+		if (*p != *q)
+			break;
+
+	/* Did not find .mgc, restore q */
+	if (p >= ext)
+		while (*q)
+			q++;
+
+	q++;
+	/* Compatibility with old code that looked in .mime */
+	if (ms->flags & MAGIC_MIME) {
+		asprintf(&buf, "%.*s.mime%s", (int)(q - fn), fn, ext);
+		if (access(buf, R_OK) != -1) {
+			ms->flags &= MAGIC_MIME_TYPE;
+			return buf;
+		}
+		free(buf);
+	}
+	asprintf(&buf, "%.*s%s", (int)(q - fn), fn, ext);
+
+	/* Compatibility with old code that looked in .mime */
+	if (strstr(p, ".mime") != NULL)
+		ms->flags &= MAGIC_MIME_TYPE;
 	return buf;
 }
 
@@ -1818,9 +2279,10 @@ swap4(uint32_t sv)
 private uint64_t
 swap8(uint64_t sv)
 {
-	uint32_t rv;
+	uint64_t rv;
 	uint8_t *s = (uint8_t *)(void *)&sv; 
 	uint8_t *d = (uint8_t *)(void *)&rv; 
+#if 0
 	d[0] = s[3];
 	d[1] = s[2];
 	d[2] = s[1];
@@ -1829,6 +2291,16 @@ swap8(uint64_t sv)
 	d[5] = s[6];
 	d[6] = s[5];
 	d[7] = s[4];
+#else
+	d[0] = s[7];
+	d[1] = s[6];
+	d[2] = s[5];
+	d[3] = s[4];
+	d[4] = s[3];
+	d[5] = s[2];
+	d[6] = s[1];
+	d[7] = s[0];
+#endif
 	return rv;
 }
 
@@ -1843,7 +2315,7 @@ bs1(struct magic *m)
 	m->in_offset = swap4((uint32_t)m->in_offset);
 	m->lineno = swap4((uint32_t)m->lineno);
 	if (IS_STRING(m->type)) {
-		m->str_count = swap4(m->str_count);
+		m->str_range = swap4(m->str_range);
 		m->str_flags = swap4(m->str_flags);
 	}
 	else {
